/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */

import Foundation
import UIKit
public class FormListeningService: EventSerivceBase {
    
    private let LOG_TAG = "BC_FORMSERVICE"
    private var _profileChanges: [ProfileChange] = []
    
    private var _eventManager: BlueConicEventFactory!
    
    private var _clickRuleMap = [UIControl: [JSON]]()
    @objc weak var delegate: UIPickerViewDelegate?
    
    public struct ProfileChange {
        var property: String
        var values: [String]
        var method: Method
        var description: String {
            return "Property: \(property), values \(values), method: \(method.description)"
        }
    }
    
    /**
     Inits everything needed for the Form Listening service
     - parameter listenerId: The id of the listener
     */
    @objc public convenience init(client: BlueConicClient, listenerId: String) {
        self.init(client, listenerId: listenerId)
    }
    
    /**
     Convenience class to apply rules. This is then transformed into a json array object my mapping each arraymember into a JSON object.
     - parameter rules: An array of Any
     */
    @objc public func applyRules(_ rules: [Any]) {
        applyRules(rules.map({ JSON($0) }))
    }
    
    /*
     Loops over all available rules and applies them when possible
     - parameter rules:  an array of rules in a JSON object
     */
    public func applyRules(_ rules: [JSON]) {
        for rule in rules {
            if let actionType = rule[PluginConstants.Listener.TAG_ACTION][PluginConstants.Listener.TAG_TYPE].string {
                if actionType == PluginConstants.Listener.TAG_ONCHANGE {
                    if let selector = rule[PluginConstants.Listener.TAG_FORMFIELD][PluginConstants.Listener.TAG_SELECTOR].string {
                        //Add something for pickerview
                        if let element = self._client.getView(selector) as? UIControl {
                            self.BCLog("Apply on change rule: \(rule)")
                            let targetMethod = #selector(FormListeningService.fieldChangeAction(_:))
                            switch element {
                            case is UISegmentedControl, is UIDatePicker, is UIStepper, is UISwitch :
                                element.addTarget(self, action: targetMethod, for: UIControl.Event.valueChanged)
                            case is UISlider:
                                if let element = element as? UISlider, element.isContinuous {
                                    element.addTarget(self, action: targetMethod, for: UIControl.Event.touchUpInside)
                                } else {
                                    element.addTarget(self, action: targetMethod, for: UIControl.Event.valueChanged)
                                }
                            case is UITextField:
                                element.addTarget(self, action: targetMethod, for: UIControl.Event.editingDidEndOnExit)
                                element.addTarget(self, action: targetMethod, for: UIControl.Event.editingDidEnd)
                                element.addTarget(self, action: targetMethod, for: UIControl.Event.valueChanged)
                            default:
                                self.BCLog("Unable to add a target to the element: \(element)")
                            }
                            self.addToClickRuleMap(element, rule: rule)
                            self.registerEvent(rule, className: NSStringFromClass(FieldChangeEvent.self))
                        } else if var _ = self._client.getView(selector) as? UIPickerView {
                            self.BCLog( "Can't listen to UIPickerView realtime, throw a FieldChangeEvent instead")
                        } else {
                            self.BCLog("Apply on value change rule: \(rule)")
                            self.registerEvent(rule, className: NSStringFromClass(UpdateValuesEvent.self))
                        }
                    }
                } else if actionType == PluginConstants.Listener.TAG_CLICKSELECTION {
                    if let selector = rule[PluginConstants.Listener.TAG_ACTION][PluginConstants.Listener.TAG_SELECTOR].string {
                        // We need to use var to addTarget
                        if let element = self._client.getView(selector) as? UIControl {
                            self.BCLog("Apply on click rule: \(rule)")
                            // map objects with their rule, since we can't set parameters with addTarget
                            element.addTarget(self, action: #selector(FormListeningService.touchAction(_:)), for: UIControl.Event.touchDown)
                            self.addToClickRuleMap(element, rule: rule)
                        } else {
                            self.BCLog("Unable to get the view with the selector: \(selector), for a click action")
                        }
                        self.registerEvent(rule, className: NSStringFromClass(ClickEvent.self))
                    }
                } else if actionType == PluginConstants.Listener.TAG_FORMSELECTION {
                    if let selector = rule[PluginConstants.Listener.TAG_ACTION][PluginConstants.Listener.TAG_SELECTOR].string {
                        
                        // We need to use var to addTarget
                        if let element = self._client.getView(selector) as? UIControl {
                            self.BCLog("Apply on form selection rule: \(rule)")
                            // map objects with their rule, since we can't set parameters with addTarget
                            element.addTarget(self, action: #selector(FormListeningService.touchAction(_:)), for: UIControl.Event.touchDown)
                            self.addToClickRuleMap(element, rule: rule)
                        } else {
                            self.BCLog("Unable to get the view with the selector: \(selector), for a form action")
                        }
                        self.registerEvent(rule, className: NSStringFromClass(FormSubmitEvent.self))
                    }
                } else {
                    self.BCLog("Appling a rule with the type '\(actionType)' is not supported", flag: Constants.Flag.WARN)
                }
            }
        }
        
        subscribeListeners()
    }
    
    /**
     Handles an event (eg click event, advanced event), which leads to points for keywords
     - parameter event: Event object
     */
    public override func handleEvent(_ event: Event) {
        self.BCLog("Handling event \(event)")
        
        if let rules = getRules(event) {
            if let formSubmitEvent = event as? FormSubmitEvent {
                for rule in rules {
                    if let profileChange = self.handleTouchEvent(rule, clickEvent: formSubmitEvent) {
                        self._profileChanges.append(profileChange)
                    }
                }
            } else if let fieldChangeEvent = event as? FieldChangeEvent {
                for rule in rules {
                    if let profileChange = self.handleFieldChangeEvent(rule, fieldChangeEvent: fieldChangeEvent) {
                        self._profileChanges.append(profileChange)
                    }
                }
            } else if let clickEvent = event as? ClickEvent {
                for rule in rules {
                    if let profileChange = self.handleTouchEvent(rule, clickEvent: clickEvent) {
                        self._profileChanges.append(profileChange)
                    }
                }
            } else if let updateValuesEvent = event as? UpdateValuesEvent {
                for rule in rules {
                    if let profileChange = self.handleUpdateValuesEvent(rule, updateValuesEvent: updateValuesEvent) {
                        self._profileChanges.append(profileChange)
                    }
                }
            }
            // save changes
            self.save()
        }
    }

    /**
     Get value from the form fields.
     */
    private func getValues(_ rule: JSON, context: [String]?) -> [String]? {
        var values: [String] = []
        if let selector = rule[PluginConstants.Listener.TAG_FORMFIELD][PluginConstants.Listener.TAG_SELECTOR].string {
            if let element = self._client.getView( selector) as? UIPickerView {
                let cont = ListenerUtil.getContent(selector, context: context)
                self.BCLog("The content from getValues: \(cont)")
                for index in 0 ..< element.numberOfComponents {
                    
                    let rowIndex = element.selectedRow(inComponent: index)
                    self.BCLog("Index: \(index) rowindex: \(rowIndex)")
                    if let rowTitle = element.delegate?.pickerView?(element, attributedTitleForRow: rowIndex, forComponent: index) {
                        values.append(rowTitle.description)
                    } else if let rowTitle = element.delegate?.pickerView?(element, titleForRow: rowIndex, forComponent: index) {
                        values.append(rowTitle)
                    }
                }
            }
            if let element = self._client.getView( selector) as? UIControl {
                if let element = element as? UIDatePicker {
                    values = [element.date.description]
                }
                if let segmentControl = element as? UISegmentedControl,
                    let newValue = segmentControl.titleForSegment(at: segmentControl.selectedSegmentIndex) {
                    values = [newValue]
                } else if let element = element as? UISwitch {
                    values = [element.isOn.description]
                } else if let element = element as? UIStepper {
                    values = [element.value.description]
                } else if let element = element as? UISlider {
                    values = [element.value.description]
                } else if let element = element as? UIButton,
                    let title = element.currentTitle {
                    values = [title]
                } else if element is UITextField {
                    values = ListenerUtil.getContent(selector, context: context)
                } else {
                    self.BCLog("No matching type")
                }
            } else {
                self.BCLog("Unable to get the view with the selector: \(selector), for a form action")
            }
        }

        // Added context as a fall back, so that developers are still able to use context elements
        // in Click- and/ or FormSubmitEvents
        return values.count > 0 ? values : context
    }
    
    private func addProfileChange(_ rule: JSON, context: [String]?) -> ProfileChange? {
        if let profileProperty: String = rule[PluginConstants.Listener.TAG_FORM_PROFILE_PROPERTY].string,
            let addSet = rule[PluginConstants.Listener.TAG_ADD_SET].method,
            let values = getValues(rule, context: context) {
            return ProfileChange(property: profileProperty, values: convertValues(rule, values: values), method: addSet)
        }
        return nil
    }
    
    /**
     The changes object only exists if there have been any changes, convenience function to check this.
     
     - returns:  bool that denotes a change has occured
     */
    @objc public func isChanged() -> Bool {
        return !self._profileChanges.isEmpty
    }
    
    /**
     Save points to BlueConic by setting the changes object to a property.
     */
    @objc public func save() {
        if self.isChanged() {
            self.BCLog("Saving")
            for profileChange: ProfileChange in self._profileChanges {
                self.saveChange(profileChange)
            }
        }
        self._profileChanges = []
        
    }
    
    private func saveChange(_ change: ProfileChange) {
        self.BCLog("Profile Change: \(change.description)")
        if change.method == Method.Set {
            self._client.setProfileValues(change.property, values: change.values)
        } else {
            self._client.addProfileValues(change.property, values: change.values)
        }
    }
    
    private func BCLog(_ log: String, flag: String = Constants.Flag.INFO) {
        BlueConicConfiguration.log(log, tag: self.LOG_TAG, flag: flag)
    }
    
    /**
     Handles an touch event rule as set in blueconic. Similar to handleEventRule, handleContentRule, etc. Unlike other rule handlers this is called when a touch action is detected
     
     - parameter rule:  A Json object describing the rule as configured in blueconic
     - parameter clickEvent: A ClickEvent
     - returns: A profile change
     */
    private func handleTouchEvent(_ rule: JSON, clickEvent: ClickEvent) -> ProfileChange? {
        if let actionSelector = rule[PluginConstants.Listener.TAG_ACTION][PluginConstants.Listener.TAG_SELECTOR].string,
            actionSelector == clickEvent.selector,
            let selector = rule[PluginConstants.Listener.TAG_FORMFIELD][PluginConstants.Listener.TAG_SELECTOR].string,
            let valueText = rule[PluginConstants.Listener.TAG_VALUE_TEXT].string,
            valueText == "value" || !ListenerUtil.getContent(selector, context: clickEvent.context).isEmpty {
            return addProfileChange(rule, context: clickEvent.context)
        }
        return nil
    }
    
    /**
     Handles the update value event. change the profile property by the values from the update values event.
     - parameter rule: A Json object describing the rule as configured in blueconic
     - parameter updateValuesEvent: the event with the profile property values.
     - returns: A profile change
     */
    private func handleUpdateValuesEvent(_ rule: JSON, updateValuesEvent: UpdateValuesEvent) -> ProfileChange? {
        if let selector = rule[PluginConstants.Listener.TAG_FORMFIELD][PluginConstants.Listener.TAG_SELECTOR].string, selector == updateValuesEvent.selector,
            let profileProperty: String = rule[PluginConstants.Listener.TAG_FORM_PROFILE_PROPERTY].string,
            let addSet = rule[PluginConstants.Listener.TAG_ADD_SET].method, updateValuesEvent.values.count > 0 {
            return ProfileChange(property: profileProperty, values: convertValues(rule, values: updateValuesEvent.values), method: addSet)
        }
        return nil
    }
    
    /**
     Handles an touch event rule as set in blueconic. Similar to handleEventRule, handleContentRule, etc. Unlike other rule handlers this is called when a touch action is detected
     
     - parameter touchRule:  A Json object describing the rule as configured in blueconic
     - parameter clickEvent:  A click event
     - returns: A profile change
     */
    private func handleFieldChangeEvent(_ rule: JSON, fieldChangeEvent: FieldChangeEvent) -> ProfileChange? {
        if let selector = rule[PluginConstants.Listener.TAG_FORMFIELD][PluginConstants.Listener.TAG_SELECTOR].string, selector == fieldChangeEvent.selector,
            let valueText = rule[PluginConstants.Listener.TAG_VALUE_TEXT].string,
            valueText == "value" || !ListenerUtil.getContent(selector, context: fieldChangeEvent.context).isEmpty {
            return addProfileChange(rule, context: fieldChangeEvent.context)
        }
        return nil
    }
    
    /**
     Respond to fieldChangeAction
     - parameter sender: the UI element that triggered the event
     */
    @objc public func fieldChangeAction(_ sender: UIControl) {
        if let touchRules = self._clickRuleMap[sender] {
            for touchRule in touchRules {
                if let selector = touchRule["formfield"]["selector"].string {
                    BlueConicEventFactory.getInstance().publish(FieldChangeEvent(selector:selector))
                    break
                }
            }
        }
    }
    
    /**
     Respond to touchAction
     - parameter sender: the UI element that triggered the event
     */
    @objc public func touchAction(_ sender: UIControl) {
        if let touchRules = self._clickRuleMap[sender] {
            for touchRule in touchRules {
                if let selector = touchRule["action"]["selector"].string {
                    BlueConicEventFactory.getInstance().publish(ClickEvent(selector:selector))
                    BlueConicEventFactory.getInstance().publish(FormSubmitEvent(selector:selector))
                    break
                }
            }
        }
    }
    
    private func addToClickRuleMap(_ element: UIControl, rule: JSON) {
        if self._clickRuleMap[element] == nil {
            self._clickRuleMap[element] = [rule]
        } else {
            var rules: [JSON] = self._clickRuleMap[element]!
            rules.append(rule)
            self._clickRuleMap[element] = rules
        }
    }
    
    public func convertValues(_ rule: JSON, values: [String]) -> [String] {
        if let mappings = rule[PluginConstants.Listener.TAG_CONVERTER][PluginConstants.Listener.TAG_MAPPINGS].array {
            return values.map({self.getConvertedValue($0, mappings: mappings)})
        }
        return values
    }
    
    private func getConvertedValue(_ value: String, mappings: [JSON]) -> String {
        for mapping in mappings {
            if let original = mapping[PluginConstants.Listener.MAPPINGS.ORIGINAL].string,
                let change = mapping[PluginConstants.Listener.MAPPINGS.CONVERTED].string,
                original == value {
                return change
            }
        }
        return value
    }
}
